import time
import io
from zipfile import ZipFile
from dataclasses import dataclass

import requests
import pandas as pd

from toargridding.metadata import Metadata, AnalysisRequestResult, Coordinates


STATION_LAT = "station_coordinates_lat"
STATION_LON = "station_coordinates_lng"
COORDS = [STATION_LAT, STATION_LON]


class AnalysisService:
    METADATA = "metadata"

    def __init__(self, stats_endpoint):
        self.stats_endpoint = stats_endpoint
        self.wait_time = 30
        self.max_response_time = 60 * 30  # 30 min

    def get_data(self, metadata: Metadata) -> AnalysisRequestResult:
        timeseries, timeseries_metadata = self.get_timeseries_and_metadata(metadata)
        coords = self.get_clean_coords(timeseries_metadata)
        timeseries = self.get_clean_timeseries(timeseries, metadata)
        return AnalysisRequestResult(timeseries, coords, metadata)

    def get_clean_coords(self, timeseries_metadata):
        coords = timeseries_metadata[COORDS]
        coords.columns = [Coordinates.latitude.name, Coordinates.longitude.name]
        valid_coords = coords.notna().all(axis=1)
        return coords[valid_coords]

    def get_clean_timeseries(self, timeseries, metadata: Metadata):
        all_na = timeseries.isna().all(axis=1)
        timeseries = timeseries[~all_na]
        timeseries = timeseries.fillna(0)
        response_timesteps = timeseries.columns
        expected_timesteps = metadata.time.as_datetime_index()
        if not (response_timesteps == expected_timesteps).all():
            raise ValueError("foo")

        # TODO maybe use cf-index here already ?
        timeseries.columns = metadata.time.as_datetime_index()
        return timeseries

    def get_timeseries_and_metadata(self, metadata: Metadata):
        query_options = self.get_query_options(metadata)
        content = self.query_timeseries_and_metadata(query_options)
        timeseries, timeseries_metadata = self.load_data(content, metadata)
        return timeseries, timeseries_metadata

    def get_query_options(self, metadata: Metadata):
        return {
            "daterange": metadata.time.daterange_option,
            "variable_id": metadata.variable.toar_id,
            "statistics": metadata.statistic,
            "sampling": metadata.time.sampling,
            "min_data_capture": 0,  # TODO check effect on NaNs
            "metadata_scheme": "basic",
            "limit": "None",
            "format": "by_statistic",
        }

    def query_timeseries_and_metadata(self, query_options, metadata) -> bytes:
        response = requests.get(self.stats_endpoint, params=query_options)
        return self.wait_for_data(response.json()["status"])  # TODO str

    def wait_for_data(self, status_endpoint):
        # TODO proper waiting
        for total_wait_time in range(0, self.max_response_time, self.wait_time):
            time.sleep(self.wait_time)

            response = requests.get(status_endpoint)  # TODO handle server errors

            if response.headers["Content-Type"] == "application/zip":
                return response.content
        else:
            raise Exception(f"Time Out after {self.max_response_time}")

    def load_data(
        self, content: bytes, metadata: Metadata
    ) -> (pd.DataFrame, pd.DataFrame):
        zip_stream = io.BytesIO(content)
        with ZipFile(zip_stream) as myzip:
            timeseries = self.extract_data(myzip, metadata.statistic)
            timeseries_metadata = self.extract_data(myzip, AnalysisService.METADATA)

        return timeseries, timeseries_metadata

    def extract_data(self, zip_file, data_file) -> pd.DataFrame:
        with zip_file.open(f"{data_file}.csv") as f:
            s_stream = io.StringIO(f.read().decode("utf-8"))
            return pd.read_csv(s_stream, comment="#", index_col=0)
